/*
 * HFS (aka MAC) filesystem/handler for MorphOS/AmigaOS
 * Author: Marek 'March' Szyprowski <marek@amiga.pl>
 */

#include <exec/types.h>
#include <dos/dosextens.h>
#include <dos/filehandler.h>
#include <devices/inputevent.h>

#include <proto/exec.h>
#include <proto/dos.h>

#include "mac_fs.h"
#include "mac_protos.h"

struct DosPacket *GetPacket(struct MsgPort *port)
{
	struct Message *msg;

	if ((msg = GetMsg(port)) != NULL)
		return (struct DosPacket *) msg->mn_Node.ln_Name;
	else
		return NULL;
}

void ProcessPackets(void)
{
	struct DosPacket *packet;

	while ((packet = GetPacket(glob->ourport)))
	{
		LONG res = DOSFALSE;
		LONG err = 0;

		switch(packet->dp_Type)
		{
			case ACTION_LOCATE_OBJECT:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				UBYTE *path = BADDR(packet->dp_Arg2);
				LONG access = packet->dp_Arg3;

				kprintf("\nGot ACTION_LOCATE_OBJECT\n");
				kprintf("\tfl = %lx fl_Key %lx ", packet->dp_Arg1, fl ? fl->fl_Key : 0);
				kprintf("type: %s\n", (packet->dp_Arg3 == EXCLUSIVE_LOCK) ? (LONG)"EXCLUSIVE" : (LONG)"SHARED");

				if ((err = TestLock(fl)))
					break;

				if (path[0] != 0)
				{
					err = LockByPath(glob->sb, fl, &path[1], path[0], access, &res);
				}
				else if (fl != NULL)
				{
					kprintf("\tCopying lock\n");
					err = LockObject(glob->sb, fl->fl_Key, fl->parid, access, &res);
				}
                else
				{
					kprintf("\tLocking root directory.\n");
					err = LockObject(glob->sb, HFS_CNID_ROOTDIR, HFS_CNID_ROOTPAR, access, &res);
				}

				break;
			}

			case ACTION_FREE_LOCK:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				kprintf("\nGot FREE_LOCK\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);

				if (fl && (err = FreeLock(fl)) == 0)
					res = DOSTRUE;
				break;
			}

			case ACTION_COPY_DIR:
			case ACTION_COPY_DIR_FH:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				kprintf("\nGot ACTION_COPY_DIR\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)))
					break;

				if (fl)
					err	= LockObject(glob->sb, fl->fl_Key, fl->parid, fl->fl_Access, &res);
				else
					err = LockObject(glob->sb, HFS_CNID_ROOTDIR, HFS_CNID_ROOTPAR, SHARED_LOCK, &res);

				break;
			}

			case ACTION_PARENT:
			case ACTION_PARENT_FH:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				kprintf("\nGot ACTION_Parent\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);
				
				if ((err = TestLock(fl)))
					break;

				if (fl && !IsRootLock(fl))
					err	= LockParent(glob->sb, fl, SHARED_LOCK, &res);
				else
					err = 0;

				break;
			}

			case ACTION_SAME_LOCK:
			{
				struct ExtFileLock *fl1 = BADDR(packet->dp_Arg1);
				struct ExtFileLock *fl2 = BADDR(packet->dp_Arg2);
				kprintf("\nGot ACTION_SAME_LOCK\n");
				err = 0;

				if (fl1 == fl2 || ((fl1->fl_Volume == fl2->fl_Volume) && fl1->fl_Key == fl2->fl_Key))
					res = DOSTRUE;

				break;
			}

			case ACTION_EXAMINE_OBJECT:
			case ACTION_EXAMINE_FH:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				struct FileInfoBlock *fib = BADDR(packet->dp_Arg2);
#ifdef __DEBUG__
				switch (packet->dp_Type)
				{
					case ACTION_EXAMINE_OBJECT:
						kprintf("\nGot ACTION_EXAMINE_OBJECT\n");
						break;
                    case ACTION_EXAMINE_FH:
						kprintf("\nGot ACTION_EXAMINE_FH\n");
						break;
                }
                kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);
#endif
				if ((err = TestLock(fl)))
					break;

				if ((err = FillFIB(glob->sb, fl->fl_Key, fl->parid, fib, FALSE)) == 0)
				{
					res = DOSTRUE;
					fl->lastcid = fl->fl_Key;
				}

				break;
			}

			case ACTION_EXAMINE_NEXT:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				struct FileInfoBlock *fib = BADDR(packet->dp_Arg2);
				kprintf("\nGot ACTION_EXAMINE_NEXT\n");
                kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)))
					break;

				if ((err = DirNextObj(glob->sb, fl, fib, FALSE)) == 0)
					res = DOSTRUE;
				break;
			}

			case ACTION_DELETE_OBJECT:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				UBYTE *path = BADDR(packet->dp_Arg2);
				kprintf("\nGot ACTION_DELETE_OBJECT\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)))
					break;
				if ((err = WriteCheck()))
					break;

				if ((err = DeleteObject(glob->sb, fl, &path[1], path[0])) == 0)
					res = DOSTRUE;
				break;
			}

			case ACTION_CREATE_DIR:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				UBYTE *path = BADDR(packet->dp_Arg2);
				kprintf("\nGot ACTION_CREATE_DIR\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)))
					break;
				if ((err = WriteCheck()))
					break;

				err = CreateDirectory(glob->sb, fl, &path[1], path[0], &res);
				break;
			}

			case ACTION_RENAME_OBJECT:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				struct ExtFileLock *fl_dest = BADDR(packet->dp_Arg3);
				UBYTE *path = BADDR(packet->dp_Arg2);
				UBYTE *path_dest = BADDR(packet->dp_Arg4);
				kprintf("\nGot ACTION_RENAME_OBJECT\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg1, fl ? fl->fl_Key : 0);
				kprintf("\tfl_dest = %lx fl_Key %lx\n", packet->dp_Arg3, fl_dest->fl_Key);

				if ((err = TestLock(fl)) || (err = TestLock(fl_dest)))
					break;
				if ((err = WriteCheck()))
					break;

				if ((err = RenameObject(glob->sb, fl, &path[1], path[0], fl_dest, &path_dest[1], path_dest[0])) == 0)
					res = DOSTRUE;
				break;
			}

			case ACTION_FH_FROM_LOCK:
			{
				struct FileHandle *fh = BADDR(packet->dp_Arg1);
				struct ExtFileLock *fl = BADDR(packet->dp_Arg2);
				kprintf("\nGot ACTION_FH_FROM_LOCK\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)))
					break;

				if (fl && (err = FileOpen(glob->sb, fl)) == 0)
				{
					fh->fh_Arg1 = (LONG)packet->dp_Arg2;
					fh->fh_Port = DOSFALSE;
					res = DOSTRUE;
				}
				else
					err = ERROR_OBJECT_WRONG_TYPE;
				break;
			}

			case ACTION_FINDINPUT:
			case ACTION_FINDUPDATE:
			case ACTION_FINDOUTPUT:
			{
				struct FileHandle *fh = BADDR(packet->dp_Arg1);
				struct ExtFileLock *fl = BADDR(packet->dp_Arg2);
				UBYTE *path = BADDR(packet->dp_Arg3);
#ifdef __DEBUG__
				switch (packet->dp_Type)
				{
                case ACTION_FINDINPUT:
					kprintf("\nGot ACTION_FINDINPUT\n");
					kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);
					break;
				case ACTION_FINDUPDATE:
					kprintf("\nGot ACTION_FINDUPDATE\n");
					kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);
					break;
				case ACTION_FINDOUTPUT:
					kprintf("\nGot ACTION_FINDOUTPUT\n");
					kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);
					break;
				default:
				}
#endif
				if ((err = TestLock(fl)))
					break;

				if (packet->dp_Type != ACTION_FINDINPUT && (err = WriteCheck(glob )))
					break;

				if ((err = CreateOpenFile(glob->sb, fl, &path[1], path[0], packet->dp_Type, fh)) == 0)
					res = DOSTRUE;
				break;
			}

			case ACTION_READ:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				APTR buffer = (APTR)packet->dp_Arg2;
				ULONG togo = packet->dp_Arg3;
				kprintf("\nGot ACTION_READ\n");
				kprintf("\tfl = %lx to read %ld bytes\n", packet->dp_Arg1, togo);

				res = -1; /* error */
				if ((err = TestLock(fl)) == 0)
				{
					if ((err = FileRead(glob->sb, fl, togo, buffer, &res)) == 0)
						break;
				}
				break;
			}

			case ACTION_WRITE:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				APTR buffer = (APTR)packet->dp_Arg2;
				ULONG togo = packet->dp_Arg3;
				kprintf("\nGot ACTION_WRITE\n");
				kprintf("\tfl = %lx to write %ld bytes\n", packet->dp_Arg1, togo);

				if ((err = TestLock(fl)) == 0 && (err = WriteCheck()) == 0)
				{
					if ((err = FileWrite(glob->sb, fl, togo, buffer, &res)) == 0)
						break;
				}
				res = -1; /* error */
				break;
			}
 
			case ACTION_SEEK:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				LONG seekoffset = packet->dp_Arg2;
				LONG mode = packet->dp_Arg3;
				kprintf("\nGot ACTION_SEEK\n");
				kprintf("\tfl = %lx move %ld type %lx\n", packet->dp_Arg1, seekoffset, mode);

				res = -1; /* error */
				if ((err = TestLock(fl)) == 0)
				{
					if ((err = FileSeek(glob->sb, fl, seekoffset, mode, &res)) == 0)
						break;
				}
                break;
			}

			case ACTION_SET_FILE_SIZE:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				LONG newsize = packet->dp_Arg2;
				LONG mode = packet->dp_Arg3;
				kprintf("\nGot ACTION_SET_FILE_SIZE\n");
				kprintf("\tfl = %lx newsize %ld type %lx\n", packet->dp_Arg1, packet->dp_Arg2, packet->dp_Arg3);

				res = -1; /* error */
				if ((err = TestLock(fl)) || (err = WriteCheck()))
					break;

				err = FileSetSize(glob->sb, fl, newsize, mode, &res);
				break;
			}

			case ACTION_END:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				kprintf("\nGot ACTION_END\n");
				kprintf("\tfl = %lx\n", packet->dp_Arg1);

				if (fl)
				{
					if ((err = FreeLock(fl)) == 0)
						res = DOSTRUE;
				}
				break;
			}

			case ACTION_IS_FILESYSTEM:
			{
				kprintf("\nGot ACTION_IS_FS\n");
				res = DOSTRUE;
				break;
			}

			case ACTION_CURRENT_VOLUME:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg1);
				kprintf("\nGot ACTION_CURRENT_VOLUME\n");

				res = (fl) ? fl->fl_Volume : ((glob->sb != NULL) ? MKBADDR(glob->sb->doslist) : NULL);
				break;
			}

			case ACTION_INFO:
			case ACTION_DISK_INFO:
			{
				struct InfoData *id;
		
				if (packet->dp_Type == ACTION_INFO)
				{
                    struct FileLock *fl = BADDR(packet->dp_Arg1);
					kprintf("\nGot ACTION_INFO\n");
					if (fl && (glob->sb == NULL || fl->fl_Volume != MKBADDR(glob->sb->doslist)))
                    {
						err = ERROR_DEVICE_NOT_MOUNTED;
						break;
					}
					id = BADDR(packet->dp_Arg2);
				}
				else
				{
					kprintf("\nGot ACTION_DISK_INFO\n");
					id = BADDR(packet->dp_Arg1);
				}
			    FillDiskInfo(id);
				res = DOSTRUE;
				break;
			}

			case ACTION_RENAME_DISK:
			{
				UBYTE *newname = BADDR(packet->dp_Arg1);
				kprintf("\nGot ACTION_RENAME_DISK\n");

				if ((err = TestLock(NULL)) != 0)
					break;

				if ((err = WriteCheck()) == 0)
				{
					if (AttemptLockDosList(LDF_VOLUMES|LDF_WRITE))
					{
						if ((err = RenameDisk(glob->sb, &newname[1], newname[0])) == 0)
							res = DOSTRUE;
						UnLockDosList(LDF_VOLUMES|LDF_WRITE);
					}
					else
					{
						kprintf("\tDosList is locked\n");
						Delay(5);
						PutMsg(glob->ourport, packet->dp_Link);
						packet = NULL;
						kprintf("Message moved to the end of the queue\n");
					}
				}
				break;
			}

			case ACTION_INHIBIT:
			{
				LONG inhibit = packet->dp_Arg1;
				kprintf("\nGot ACTION_INHIBIT\n");
				if (inhibit == DOSTRUE)
				{
					glob->disk_inhibited++;
					if (glob->disk_inhibited == 1 && glob->sb)
						DoDiskRemove();
				}
				else
				{
					glob->disk_inhibited--;
					if (glob->disk_inhibited == 0)
					   ProcessDiskChange();
				}
				res = DOSTRUE;
				break;
			}

			case ACTION_DIE:
			{
				kprintf("\nGot ACTION_DIE\n");
				if (glob->sblist != NULL || (glob->sb != NULL && (glob->sb->doslist->dol_misc.dol_volume.dol_LockList != NULL)))
				{
					kprintf("\tThere are some locks/volumes left. Shutting down is not possible\n");
					err = ERROR_OBJECT_IN_USE;
					break;
				}
				kprintf("\tNo locks pending. Shutting down the handler\n");

			 /* the following code works, because it will put diskchange msg in our msg queue and process it in next loop call */
				SendDiskChangePacket(ACTION_DOSNODE_KILL, NULL);
				DoDiskRemove();
				glob->quit = TRUE;
				glob->disk_inhibited += 1;

				res = DOSTRUE;
				break;
			}

            case ACTION_GET_DISK_FSSM:
			{
				kprintf("\nGot ACTION_GET_DISK_FSSM\n");
				res = (ULONG) glob->fssm;
				break;
			}
 
			case ACTION_FREE_DISK_FSSM:
			{
				kprintf("\nGot ACTION_FREE_DISK_FSSM\n");
				res = DOSTRUE;
				break;
 
			}

			case ACTION_DISK_CHANGE: /* internal */
			{
				struct DosList *vol = (struct DosList *)packet->dp_Arg2;
				ULONG type = packet->dp_Arg3;
				kprintf("\nGot ACTION_DISK_CHANGE\n");

				if (packet->dp_Arg1 == ID_HFS_DISK) /* security check */
				{
					if (AttemptLockDosList(LDF_VOLUMES|LDF_WRITE))
					{
						if (type == ACTION_VOLUME_ADD)
						{
							AddDosEntry(vol);
							UnLockDosList(LDF_VOLUMES|LDF_WRITE);
							SendEvent(IECLASS_DISKINSERTED);
							kprintf("\tVolume added successfuly\n");
						}
						else if (type == ACTION_VOLUME_REMOVE)
						{
							RemDosEntry(vol);
							FS_FreeMem(vol);
							UnLockDosList(LDF_VOLUMES|LDF_WRITE);
							SendEvent(IECLASS_DISKREMOVED);
							kprintf("\tVolume removed successfuly.\n");
						}
						else if (type == ACTION_DOSENVEC_UPDATE)
						{
							struct FileSysStartupMsg *dev_fssm = BADDR(glob->devnode->dol_misc.dol_handler.dol_Startup);
							struct DosEnvec *dev_de = BADDR(dev_fssm->fssm_Environ);
							struct DosEnvec *de = BADDR(glob->fssm->fssm_Environ);

							memcpy(dev_de, de, sizeof(*de));
							UnLockDosList(LDF_VOLUMES|LDF_WRITE);
							kprintf("\tDosEnvec successfuly updated.\n");
						}
						else if (type == ACTION_DOSNODE_KILL)
						{
							glob->devnode->dol_Task = NULL;
							UnLockDosList(LDF_VOLUMES|LDF_WRITE);
							kprintf("\tDosNode successfuly cleared.\n");
						}
						FreeDosObject(DOS_STDPKT, packet); /* cleanup */
						packet = NULL;
						kprintf("Packet destroyed\n");
					}
					else
					{
						kprintf("\tDosList is locked\n");
						Delay(5);
						PutMsg(glob->ourport, packet->dp_Link);
						packet = NULL;
						kprintf("Message moved to the end of the queue\n");
					}
				}
				else
					err = ERROR_OBJECT_WRONG_TYPE;
				break;
			}

			case ACTION_CHANGE_MODE:
			{
				struct ExtFileLock *fl = 0;
				kprintf("\nGot ACTION_CHANGE_MODE\n");
				kprintf("\ttype %lx\n", packet->dp_Arg1);

				if (packet->dp_Arg1 == CHANGE_FH)
					fl = BADDR(((struct FileHandle *)BADDR(packet->dp_Arg2))->fh_Arg1);
				else if (packet->dp_Arg1 == CHANGE_LOCK)
					fl = BADDR(packet->dp_Arg2);
				else
					err = ERROR_OBJECT_WRONG_TYPE;

				kprintf("\tdp_Arg2 = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);

				if(err == 0 && fl != NULL)
				{
					if(fl->fl_Access != EXCLUSIVE_LOCK && packet->dp_Arg3 == EXCLUSIVE_LOCK) /* SHARED -> EXCLUSIVE */
					{
						if (!IsRootLock(fl))
                        {
							LONG fl_Key = fl->fl_Key;
							fl->fl_Key = 0;  /* this makes sure that our lock is not taken into account by IsLockable() */

							if((err = IsLockable(glob->sb, fl_Key, EXCLUSIVE_LOCK)) == 0)
							{
								fl->fl_Access = EXCLUSIVE_LOCK;
								res = DOSTRUE;
							}
							fl->fl_Key = fl_Key; /* restore */
                        }
                        else
							err = ERROR_OBJECT_IN_USE;
					}
					else /* EXCLUSIVE -> SHARED or no change */
					{
						fl->fl_Access = packet->dp_Arg3;
						res = DOSTRUE;
					}
				}
				else
					err = ERROR_OBJECT_WRONG_TYPE;

				break;
			}

			case ACTION_FLUSH:
			{
				kprintf("\nGot ACTION_FLUSH\n");
				if ((err = TestLock(NULL)) == 0)
				{
					hfs_flushall();
					err = 0;
					res = DOSTRUE;
				}
				break;
			}

			case ACTION_WRITE_PROTECT:
			{
				kprintf("\nGot ACTION_WRITE_PROTECT\n");
				if ((err = TestLock(NULL)) != 0)
					break;

				if (packet->dp_Arg1 == DOSTRUE)
				{
					glob->sb->lock_key = packet->dp_Arg2;
					glob->sb->read_only = TRUE;
					res = DOSTRUE;
					kprintf("\tDisk write protected\n");
				}
				else if (packet->dp_Arg1 == DOSFALSE && (glob->sb->lock_key == 0 || glob->sb->lock_key == packet->dp_Arg2))
				{
					glob->sb->lock_key = 0;
					glob->sb->read_only = FALSE;
					res = DOSTRUE;
					kprintf("\tDisk write-proction removed\n");
				}
				else
					err = ERROR_OBJECT_WRONG_TYPE;
				break;
            }

			case ACTION_FORMAT:
			{
				UBYTE *name = BADDR(packet->dp_Arg1);
				kprintf("\nGot ACTION_FORMAT\n");
				if (!glob->disk_inhibited)
                {
					err = ERROR_OBJECT_IN_USE;
					break;
				}
				if ((err = FormatDisk(&name[1], name[0])) == 0)
					res = DOSTRUE;

				break;
			}

			case ACTION_SET_DATE:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg2);
				UBYTE *path = BADDR(packet->dp_Arg3);
				kprintf("\nGot ACTION_SET_DATE\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)) || (err = WriteCheck()))
					break;
 
				if ((err = SetDate(glob->sb, fl, &path[1], path[0], (struct DateStamp *)packet->dp_Arg4)) == 0)
					res = DOSTRUE;
				break;
			}
			
			case ACTION_SET_PROTECT:               /* dummies */
			case ACTION_SET_COMMENT:  
			case ACTION_SET_OWNER:
			{
				kprintf("\nGot file attribute change packet, not supported yet\n");
				res = DOSTRUE;
				break;
			}

#if 0
			case ACTION_GET_EXT_ATTR:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg2);
				UBYTE *path = BADDR(packet->dp_Arg3);
				kprintf("\nGot ACTION_GET_EXT_ATTR\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)))
					break;

				if ((err = GetExtAttr(glob->sb, fl, &path[1], path[0], (struct ExtAttr *)packet->dp_Arg4)) == 0)
					res = DOSTRUE;
				break;
			}

			case ACTION_SET_EXT_ATTR:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg2);
				UBYTE *path = BADDR(packet->dp_Arg3);
				kprintf("\nGot ACTION_SET_EXT_ATTR\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)) || (err = WriteCheck()))
					break;
 
				if ((err = SetExtAttr(glob->sb, fl, &path[1], path[0], (struct ExtAttr *)packet->dp_Arg4)) == 0)
					res = DOSTRUE;
				break;
			}
#endif

			case ACTION_SET_MAC_BOOTABLE:
			{
				struct ExtFileLock *fl = BADDR(packet->dp_Arg2);
				kprintf("\nGot ACTION_BLESS\n");
				kprintf("\tfl = %lx fl_Key %lx\n", packet->dp_Arg2, fl ? fl->fl_Key : 0);

				if ((err = TestLock(fl)))
					break;

				if ((err = SetMacBootable(glob->sb, fl)) == 0)
					res = DOSTRUE;
				break;
			}

			case ACTION_READ_LINK:
			{
				kprintf("\nGot ACTION_READ_LINK\n");
				/* Autodoc and GuruBook are wrong, -1 and -2 are used to indicate failure. - Piru */
				res = -1;
				err = ERROR_ACTION_NOT_KNOWN;
				break;
			}

			default:
				kprintf("\nGot UNKNOWN %ld\n", packet->dp_Type);
				err = ERROR_ACTION_NOT_KNOWN;
				break;
		}
	
		if (packet)
		{
			kprintf("Returning packet: %lx %lx\n", res, err);
			ReplyPkt(packet, res, err);
		}
	}
}

